#include "main.h"

//#define DEBUG
//#define DEBUG_NOISY
#define SMARTWAIT

char RinPath[MAX_PATH];
char RomPath[MAX_PATH];
char RomName[MAX_PATH];
char SavePath[MAX_PATH];
SETTING setting;

// ----------------------------------------------------------------------------------------
int exit_callback(void)
{
	char *p;

// Cleanup the games resources etc (if required)
	if (rom_get_loaded() && rom_has_battery())
		save_sram(get_sram(), rom_get_info()->ram_size);
//	save_config();

// Exit game
	sceKernelExitGame();

	return 0;
}

#define POWER_CB_POWER		0x80000000
#define POWER_CB_HOLDON		0x40000000
#define POWER_CB_STANDBY	0x00080000
#define POWER_CB_RESCOMP	0x00040000
#define POWER_CB_RESUME		0x00020000
#define POWER_CB_SUSPEND	0x00010000
#define POWER_CB_EXT		0x00001000
#define POWER_CB_BATLOW		0x00000100
#define POWER_CB_BATTERY	0x00000080
#define POWER_CB_BATTPOWER	0x0000007F
void power_callback(int unknown, int pwrflags)
{
	char *p;

	// Combine pwrflags and the above defined masks
	if(pwrflags & POWER_CB_POWER){
		if (rom_get_loaded() && rom_has_battery())
			save_sram(get_sram(), rom_get_info()->ram_size);
//		save_config();
	}
}

// Thread to create the callbacks and then begin polling
int CallbackThread(void *arg)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback);
	SetExitCallback(cbid);
	cbid = sceKernelCreateCallback("Power Callback", power_callback);
	PowerSetCallback(0, cbid);

	KernelPollCallbacks();
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

// ----------------------------------------------------------------------------------------

extern int skip, now_frame;
int xmain(int argc, char *argv)
{
	unsigned long curclock, lastdraw, lastdraw2, lastdraw3, lastdraw4;
	unsigned long waitfc=0;
	int datalength = 44100*16/8*2*BUF_LEN/1000;
	int fd, romsize, ramsize;
	int line;
	int i;
	int bPad=0;
	int bSave=0;
	char *p, savedir[MAX_PATH];

	pgMain(argc, argv);

	_strcpy(RinPath, argv);
	p = _strrchr(RinPath, '/');
	*++p = 0;

	SetupCallbacks();
	pgInit();
	pgScreenFrame(2,0);
	sceCtrlInit(0);
	wavoutInit();

	load_config();
	load_menu_bg();
	bgbright_change();
	if(setting.sound)
		//wavout_snd0_wavinfo=&wavinfo_bg;
		wavout_enable=1;

	_strcpy(savedir,RinPath);
	_strcat(savedir,"SAVE/");
	sceIoMkdir(savedir,0777);
	
	gb_init();
	_strcpy(LastPath,RinPath);
	FilerMsg[0]=0;
	for(;;){
		while(!getFilePath(RomPath))
			;
		// w肵t@C[hB by ruka
		romsize = load_rom(RomPath);
		if (!romsize) continue;
		
		p = _strrchr(RomPath,'/');
		_strcpy(RomName, p+1);

		ramsize = load_sram();
		if(ramsize) bSave = 1;
		
		if(gb_load_rom(rom_image, romsize, bSave?sram_space:0, bSave?ramsize:0))
			break;
	}

	pgFillvram(0);
	pgScreenFlipV();
	pgFillvram(0);
	pgScreenFlipV();

#ifdef DEBUG
	unsigned long lasttick;
	unsigned long framecount;
	unsigned long lastclock;
	lasttick=paddata.frame;
#endif

	lastdraw=sceKernelLibcClock();
#ifdef SMARTWAIT
	lastdraw2=lastdraw-16666;
	lastdraw3=lastdraw2-16666;
	lastdraw4=lastdraw3-16666;
#endif

	for(;;) {
#ifdef SMARTWAIT
#else
		unsigned long t;
		t=sceKernelLibcClock();
		if(t<lastdraw) lastdraw=t;
#endif

		for(line=0; line<154; line++)
			gb_run();

#if defined(SMARTWAIT)||defined(DEBUG)
		curclock=sceKernelLibcClock();
#endif

#ifdef DEBUG
		framecount++;
		if (framecount>=60) {
			unsigned long l;

			//t[[g̎wWB60t[ł(usec)PUiŕ\B
			//tt[0x000f4240ƂȂA傫ƒxƂɂȂB𑜓x̂͊فB - LCK
//			unsigned long curtick=paddata.frame;
			framecount=0;
#ifdef DEBUG_NOISY
			pgcLocate(50,0);
			pgcPuthex8(curclock-lasttick);
#endif
			lasttick=curclock;

#ifdef DEBUG_NOISY
			pgcLocate(50,2);
			pgcPuthex8(c_regs_PC);
			l=(cpu_read_direct(c_regs_PC)<<24)+(cpu_read_direct(c_regs_PC+1)<<16)+(cpu_read_direct(c_regs_PC+2)<<8)+(cpu_read_direct(c_regs_PC+3));
			pgcLocate(50,3);
			pgcPuthex8(l);

			unsigned long curclock=cpu_get_clock();
			pgcLocate(50,5);
			pgcPuthex8(curclock-lastclock);
			lastclock=curclock;

			pgcLocate(50,25);
			pgcPuthex8(g_regs.IF);
			pgcLocate(50,26);
			pgcPuthex8(g_regs.IE);
			extern byte c_regs_I;
			pgcLocate(50,27);
			pgcPuthex8(c_regs_I);
			
			// kmg
			pgcLocate(2, 2);
			pgcPuthex8(paddata.analog[CTRL_ANALOG_X]);
			pgcLocate(2, 3);
			pgcPuthex8(paddata.analog[CTRL_ANALOG_Y]);
#endif
		}
#endif

#ifdef SMARTWAIT
		if (setting.vsync) {
			pgWaitV();
		} else if (setting.wait) {
			while((curclock-lastdraw)<10000) curclock=sceKernelLibcClock();
			if (((waitfc++)&3)==0) while((curclock-lastdraw4)<16666*4) curclock=sceKernelLibcClock();
		}
		lastdraw4 = lastdraw3;
		lastdraw3 = lastdraw2;
		lastdraw2 = lastdraw;
		lastdraw = curclock;
#else
//]̃[`
		if(setting.vsync){
			pgWaitV();
			lastdraw = sceKernelLibcClock();
		}else if(setting.wait){
			while((sceKernelLibcClock()-lastdraw)<16666)
				;
			lastdraw = sceKernelLibcClock();
		}
#endif

		if(now_frame==0) pgScreenFlip();

		if(paddata.buttons && paddata.buttons==setting.key_config[6]){
			wavout_enable=0;
			rin_menu();
			if(setting.sound)
				wavout_enable=1;
		}
	}

	return 0;
}
